// see:   http://llvm.org/docs/CodingStandards.html#include-style
#include "geom.h"

#include <algorithm>
#include <cassert>
#include <fstream>
#include <iostream>
#include <list>
#include <string>
#include <vector>
using namespace std;


//  (0,1) * (0,1) is divided into procx * procy subdomains
//  numbered row by row,
//  according to myid, the array neigh contains the id's
//  of the neighbors at South, East, North, West

//  IN:  	myid	  process id
//		    procx	  number of subdomains in x-direction
//          procy	  number of subdomains in x-direction
//  OUT     neigh     id's of neighbors
//		    color     black/white coloring (check board) of subdomains

void IniGeom(int const myid, int const procx, int const procy, int neigh[], int &color)
{
    const int ky = myid / procx;
    const int kx = myid % procy;	//    MOD(myid,procx)

    neigh[0] = (ky == 0)       ?  -1 : myid - procx;    //   South
    neigh[1] = (kx == procx - 1) ?  -1 : myid + 1;      //   East
    neigh[2] = (ky == procy - 1) ?  -1 : myid + procx;  //   North
    neigh[3] = (kx == 0)       ?  -1 : myid - 1;        //   West

    color = (kx + ky) & 1 ;
    return;
}

//   (0,1) * (0,1) is divided into procx * procy subdomains
//   numbered row by row,
//   according to myid, the coordinates of the lower left corner
//   (xl,yb) and the upper right corner (xr,yt) are generated
//
//	IN:  	myid	  process id
//		    procx	  number of subdomains in x-direction
//		    procy	  number of subdomains in x-direction
//	OUT:    (xl,yb)   coordinates of the lower left corner
//		    (xr,yt)   coordinates the upper right corner

void IniCoord(int const myid, int const procx, int const procy,
              double &xl, double &xr, double &yb, double &yt)
{
    const int kx = myid % procx;
    const int ky = myid / procx;
    const double hx = 1. / procx;
    const double hy = 1. / procy;
    /* subdomain - part of unit square */
    xl = kx * hx;                      /* left    */
    xr = (kx + 1) * hx;                /*  right  */
    yb = ky * hy;                      /*  bottom */
    yt = (ky + 1) * hy;                /*  top */

    return;
}

void GetCoordsInRectangle(int const nx, int const ny,
                          double const xl, double const xr, double const yb, double const yt,
                          double xc[])
{
    const double hx = (xr - xl) / nx,
                 hy = (yt - yb) / ny;

    int k = 0;
    for (int j = 0; j <= ny; ++j)
    {
        const double y0 = yb + j * hy;
        for (int i = 0; i <= nx; ++i, k += 2)
        {
            xc[k  ] = xl + i * hx;
            xc[k + 1] = y0;
        }
    }

    return;
}

void GetConnectivityInRectangle(int const nx, int const ny, int ia[])
{
    const int dx = nx + 1;
    int k  = 0;
    int l  = 0;
    for (int j = 0; j < ny; ++j, ++k)
    {
        for (int i = 0; i < nx; ++i, ++k)
        {
            ia[l  ] = k;
            ia[l + 1] = k + 1;
            ia[l + 2] = k + dx + 1;
            l += 3;
            ia[l  ] = k;
            ia[l + 1] = k + dx;
            ia[l + 2] = k + dx + 1;
            l += 3;
        }
    }
    return;
}

//  Copies the values of w corresponding to the boundary
//  South (ib==1), East (ib==2), North (ib==3), West (ib==4)

void GetBound(int const ib, int const nx, int const ny, double const w[], double s[])
{
    const int //dx = 1,
    dy = nx + 1,
    bl  = 0,
    br  = nx,
    tl  = ny * (nx + 1),
    tr  = (ny + 1) * (nx + 1) - 1;
    switch (ib)
    {
        case 1:
        {
            for (int i = bl, j = 0; i <= br; ++i, ++j)
                s[j] = w[i];
            break;
        }
        case 3:
        {
            for (int i = tl, j = 0; i <= tr; ++i, ++j)
                s[j] = w[i];
            break;
        }
        case 4:
        {
            for (int i = bl, j = 0; i <= tl; i += dy, ++j)
                s[j] = w[i];
            break;
        }
        case 2:
        {
            for (int i = br, j = 0; i <= tr; i += dy, ++j)
                s[j] = w[i];
            break;
        }
        default:
        {
            cout << endl << "Wrong parameter  ib in " << __FILE__ << ":" << __LINE__ << endl;
        }
    }
    return;
}

// ----------------------------------------------------------------------------------------------------------
// Computes w:  = w + s at nodes on  the boundary
// South (ib == 1), East (ib == 2), North (ib == 3), West (ib == 4)

void AddBound(int const ib, int const nx, int const ny, double w[], double const s[])
{
    int const dy = nx + 1,
              bl  = 0,
              br  = nx,
              tl  = ny * (nx + 1),
              tr  = (ny + 1) * (nx + 1) - 1;
    switch (ib)
    {
        case 1:
        {
            for (int i = bl, j = 0; i <= br; ++i, ++j)
                w[i] += s[j];
            break;
        }
        case 3:
        {
            for (int i = tl, j = 0; i <= tr; ++i, ++j)
                w[i] += s[j];
            break;
        }
        case 4:
        {
            for (int i = bl, j = 0; i <= tl; i += dy, ++j)
                w[i] += s[j];
            break;
        }
        case 2:
        {
            for (int i = br, j = 0; i <= tr; i += dy, ++j)
                w[i] += s[j];
            break;
        }
        default:
        {
            cout << endl << "Wrong parameter  ib in " << __FILE__ << ":" << __LINE__ << endl;
        }
    }
    return;
}

//----------------------------------------------------------------------

void SaveVectorP(int const myid, string const &name, double const u[],
                 int const nx, int const ny,
                 double const xl, double const xr, double const yb, double const yt)
{
//  nnode = (nx+1)*(ny+1);
//  xc    = (double *) malloc(2*nnode*sizeof(double));  /* coordinates	  */
    auto *xc = new double [2 * (nx + 1) * (ny + 1)];
    GetCoordsInRectangle(nx, ny, xl, xr, yb, yt, xc);

//  construct the file name for subdomain myid
    const string tmp( std::to_string(myid / 100) + to_string((myid % 100) / 10) + to_string(myid % 10) );

    const string namep = name + "." + tmp;
    ofstream ff(namep.c_str());
    ff.precision(6);
    ff.setf(ios::fixed, ios::floatfield);

    int k = 0;
    for (int j = 0; j <= ny; ++j)
    {
        for (int i = 0; i <= nx; ++i, ++k)
            ff << xc[2 * k + 0] << "   " << xc[2 * k + 1] << "   " << u[k] << endl;
        ff << endl;
    }

    ff.close();
    delete [] xc;
    return;
}

//----------------------------------------------------------------------

void GetMesh(const int nx, const int ny, const double xl, const double xr, const double yb, const double yt,
             int &nnode, double *&xc, int &nelem, int *&ia)
{
    nnode = (nx + 1) * (ny + 1);
    nelem = 2 * nx * ny;

    xc = new double [2 * nnode];		/* coordinates	  */
    GetCoordsInRectangle(nx, ny, xl, xr, yb, yt, xc);

    ia = new int [3 * nelem]; 			/* connectivity matrix */
    GetConnectivityInRectangle(nx, ny, ia);

    return;
}

//----------------------------------------------------------------------

void DebugMesh(const int nnode, const double xc[], const int nelem, const int ndof_e, const int ia[])
{
    cout << "\n ############### Debug  M E S H  ###################\n";
    cout << "\n ...............    Coords       ...................\n";
    for (int k = 0; k < nnode; ++k)
    {
        cout << k << " : " << xc[2 * k] << "  " << xc[2 * k + 1] << endl;
    }
    cout << "\n ...............    Elems        ...................\n";
    for (int k = 0; k < nelem; ++k)
    {
        cout << k << " : ";
        for (int i = 0; i < ndof_e; ++i )
            cout << ia[ndof_e * k + i] << "  ";
        cout << endl;
    }
    return;
}


